"""
Wrapper classes for providing a minimal consistent interface to cheminformatics toolkits

Currently supported toolkits:

* The `OpenEye Toolkit <https://docs.eyesopen.com/toolkits/python/quickstart-python/index.html>`_
* The `RDKit <http://www.rdkit.org/>`_
* `AmberTools <http://ambermd.org/AmberTools.php>`_

.. todo::

   * Add checks at the beginning of each toolkit method call to make sure toolkit is licened
   * Switch toolkit methods to object methods instead of static methods
   * Should this be under ``openff.toolkit.utils.toolkits`` or ``openff.toolkit.toolkits``?
   * Add singleton global toolkit registry that registers all available toolkits by default
        when this file is imported
   * Add description fields for each toolkit wrapper
   * Eliminate global variables in favor of a singleton pattern
   * Change global variables from _INSTALLED to _AVAILABLE

"""

__all__ = (
    "ALLOWED_AROMATICITY_MODELS",
    "ALLOWED_CHARGE_MODELS",
    "ALLOWED_FRACTIONAL_BOND_ORDER_MODELS",
    "AMBERTOOLS_AVAILABLE",
    "BASIC_CHEMINFORMATICS_TOOLKITS",
    "DEFAULT_AROMATICITY_MODEL",
    "DEFAULT_CHARGE_MODEL",
    "DEFAULT_FRACTIONAL_BOND_ORDER_MODEL",
    "GLOBAL_TOOLKIT_REGISTRY",
    "OPENEYE_AVAILABLE",
    "RDKIT_AVAILABLE",
    "AmberToolsToolkitWrapper",
    "AntechamberNotFoundError",
    "BadMoleculeAssumptionError",
    "BuiltInToolkitWrapper",
    "ChargeCalculationError",
    "ChargeMethodUnavailableError",
    "GAFFAtomTypeWarning",
    "IncompatibleUnitError",
    "IncorrectNumConformersError",
    "IncorrectNumConformersWarning",
    "InvalidIUPACNameError",
    "InvalidToolkitError",
    "InvalidToolkitRegistryError",
    "LicenseError",
    "MissingPackageError",
    "NAGLToolkitWrapper",
    "OpenEyeToolkitWrapper",
    "RDKitToolkitWrapper",
    "SMILESParseError",
    "ToolkitRegistry",
    "ToolkitUnavailableException",
    "ToolkitWrapper",
    "UndefinedStereochemistryError",
)


import logging

from openff.toolkit.utils.ambertools_wrapper import AmberToolsToolkitWrapper
from openff.toolkit.utils.base_wrapper import ToolkitWrapper
from openff.toolkit.utils.builtin_wrapper import BuiltInToolkitWrapper
from openff.toolkit.utils.constants import (
    ALLOWED_AROMATICITY_MODELS,
    ALLOWED_CHARGE_MODELS,
    ALLOWED_FRACTIONAL_BOND_ORDER_MODELS,
    DEFAULT_AROMATICITY_MODEL,
    DEFAULT_CHARGE_MODEL,
    DEFAULT_FRACTIONAL_BOND_ORDER_MODEL,
)
from openff.toolkit.utils.exceptions import (
    AntechamberNotFoundError,
    BadMoleculeAssumptionError,
    ChargeCalculationError,
    ChargeMethodUnavailableError,
    GAFFAtomTypeWarning,
    IncompatibleUnitError,
    IncorrectNumConformersError,
    IncorrectNumConformersWarning,
    InvalidIUPACNameError,
    InvalidToolkitError,
    InvalidToolkitRegistryError,
    LicenseError,
    MissingPackageError,
    SMILESParseError,
    ToolkitUnavailableException,
    UndefinedStereochemistryError,
)
from openff.toolkit.utils.nagl_wrapper import NAGLToolkitWrapper
from openff.toolkit.utils.openeye_wrapper import OpenEyeToolkitWrapper
from openff.toolkit.utils.rdkit_wrapper import RDKitToolkitWrapper
from openff.toolkit.utils.toolkit_registry import (
    ToolkitRegistry,
    toolkit_registry_manager,
)

logger = logging.getLogger(__name__)


# Create global toolkit registry, where all available toolkits are registered
GLOBAL_TOOLKIT_REGISTRY = ToolkitRegistry(
    toolkit_precedence=[
        NAGLToolkitWrapper,
        OpenEyeToolkitWrapper,
        RDKitToolkitWrapper,
        AmberToolsToolkitWrapper,
        BuiltInToolkitWrapper,
    ],
    exception_if_unavailable=False,
)
"""
The toolkit registry used by default when no registry or wrapper is specified.

A toolkit registry is a list of toolkit wrappers used to provide functionality
not implemented by the OpenFF Toolkit itself. Any given functionality is
requested from each toolkit in order until one is found that provides it. For
more information, see :py:class:`ToolkitRegistry`.

To temporarily modify the ``GLOBAL_TOOLKIT_REGISTRY``,
see :py:func:`toolkit_registry_manager`.

Examples
========

Check the default toolkit precedence order:

>>> from openff.toolkit import GLOBAL_TOOLKIT_REGISTRY
>>> GLOBAL_TOOLKIT_REGISTRY.registered_toolkits() # doctest: +ELLIPSIS
[...]

Temporarily change the global toolkit registry:

>>> from openff.toolkit import ToolkitRegistry, RDKitToolkitWrapper
>>> from openff.toolkit.utils import toolkit_registry_manager
>>> with toolkit_registry_manager(ToolkitRegistry([RDKitToolkitWrapper]))):
>>>     GLOBAL_TOOLKIT_REGISTRY
<ToolkitRegistry containing The RDKit>
>>>     mol = Molecule.from_smiles("CCO")
>>> # when the context manager ends, the default registry is back to normal
>>> GLOBAL_TOOLKIT_REGISTRY
<ToolkitRegistry containing The RDKit, AmberTools, Built-in Toolkit>

To change the value of ``GLOBAL_TOOLKIT_REGISTRY`` for the remainder of the
Python process, individual toolkits can be added or removed with
the :py:meth:`ToolkitRegistry.deregister_toolkit`
and :py:meth:`ToolkitRegistry.register_toolkit` methods. This can be useful for
debugging and exploring subtly different behavior between toolkit wrappers:

>>> print(len(GLOBAL_TOOLKIT_REGISTRY.registered_toolkits))
4
>>> GLOBAL_TOOLKIT_REGISTRY.deregister_toolkit(RDKitToolkitWrapper)
>>> print(len(GLOBAL_TOOLKIT_REGISTRY.registered_toolkits))
3
>>> GLOBAL_TOOLKIT_REGISTRY.register_toolkit(RDKitToolkitWrapper)
>>> print(len(GLOBAL_TOOLKIT_REGISTRY.registered_toolkits))
4

Note that as with other global attributes in Python, assigning a new toolkit
registry to ``GLOBAL_TOOLKIT_REGISTRY`` is quite difficult to get right and very
likely to fail silently - we recommend modifying the existing value instead in
most cases.

See Also
========
toolkit_registry_manager, ToolkitRegistry, ToolkitWrapper
"""


OPENEYE_AVAILABLE = False
RDKIT_AVAILABLE = False
AMBERTOOLS_AVAILABLE = False

with toolkit_registry_manager(
    ToolkitRegistry(
        toolkit_precedence=[
            OpenEyeToolkitWrapper,
            RDKitToolkitWrapper,
            AmberToolsToolkitWrapper,
        ],
        exception_if_unavailable=False,
    )
):
    for toolkit in GLOBAL_TOOLKIT_REGISTRY.registered_toolkits:
        if type(toolkit) is OpenEyeToolkitWrapper:
            OPENEYE_AVAILABLE = True
        elif type(toolkit) is RDKitToolkitWrapper:
            RDKIT_AVAILABLE = True
        elif type(toolkit) is AmberToolsToolkitWrapper:
            AMBERTOOLS_AVAILABLE = True


# Define basic toolkits that handle essential file I/O

BASIC_CHEMINFORMATICS_TOOLKITS = [RDKitToolkitWrapper, OpenEyeToolkitWrapper]

# Ensure we have at least one basic toolkit
any_toolkits = False
for tk in GLOBAL_TOOLKIT_REGISTRY.registered_toolkits:
    if type(tk) in BASIC_CHEMINFORMATICS_TOOLKITS:
        if tk.is_available():
            any_toolkits = True
            break

if not any_toolkits:  # pragma: no cover
    from openff.toolkit.utils.utils import all_subclasses

    msg = "WARNING: No basic cheminformatics toolkits are available.\n"
    msg += "At least one basic toolkit is required to handle SMARTS matching and file I/O. \n"
    msg += "Please install at least one of the following basic toolkits:\n"
    for wrapper in all_subclasses(ToolkitWrapper):
        if wrapper.toolkit_name is not None:
            msg += f"{wrapper._toolkit_name} : {wrapper._toolkit_installation_instructions}\n"
    # TODO: Make this a warning!?
    print(msg)



"""
Tests for callback utility classes and functions.

"""

import pytest

from openff.toolkit.utils.callback import Callbackable, callback_method
from openff.toolkit.utils.exceptions import CallbackRegistrationError


class CallHistory:
    """Used to keep track of the order in which callbacks and methods are called by Callbackable."""

    history = None

    def reset_history(self):
        CallHistory.history = []

    @classmethod
    def add_history_entry(cls, name, *args, **kwargs):
        # Store args and kwargs in the history only if they are given.
        history_entry = []
        if len(args) != 0:
            history_entry.append(args)
        if len(kwargs) != 0:
            history_entry.append(kwargs)

        if len(history_entry) == 0:
            cls.history.append(name)
        else:
            cls.history.append([name, *history_entry])

    def instance_callback(self, callbackable, event_name, *args, **kwargs):
        assert isinstance(self, object)
        CallHistory.add_history_entry("callback_" + event_name, *args, **kwargs)

    @classmethod
    def class_callback(cls, callbackable, event_name, *args, **kwargs):
        assert isinstance(cls, type)
        CallHistory.add_history_entry("callback_" + event_name, *args, **kwargs)

    @staticmethod
    def static_callback(callbackable, event_name, *args, **kwargs):
        CallHistory.add_history_entry("callback_" + event_name, *args, **kwargs)


call_history = CallHistory()


class TestCallbackable:
    """Test suite for the Callbackable base class."""

    class MyCallbackable(Callbackable):
        @callback_method
        def instance_method(self, *args, **kwargs):
            CallHistory.add_history_entry("instance_method", *args, **kwargs)

        @callback_method
        def __iadd__(self, other):
            CallHistory.add_history_entry("__iadd__", other)

        @callback_method(events=["event1"])
        def event_method1(self, *args, **kwargs):
            CallHistory.add_history_entry("event_method1", *args, **kwargs)

        @callback_method(events=["event1", "event2"])
        def event_method2(self, *args, **kwargs):
            CallHistory.add_history_entry("event_method2", *args, **kwargs)

    def check_method_call_order(self, callbackable, event_name, event_sequence, *args, **kwargs):
        """Check that callback and methods/attributes are invoked in the correct order.

        This also formats the history correctly if args and kwargs are given.
        """
        # Modify the expected history if args and kwargs are given.
        if len(args) == 0 and len(kwargs) == 0:
            expected_history = event_sequence
        else:
            expected_history = [[event_name] for event_name in event_sequence]
        if len(args) != 0:
            for event in expected_history:
                event.append(args)
        if len(kwargs) != 0:
            for event in expected_history:
                event.append(kwargs)

        # Reset history and verify that the calls are in the correct order.
        call_history.reset_history()
        getattr(callbackable, event_name)(*args, **kwargs)
        assert call_history.history == expected_history

    @pytest.mark.parametrize("event_name", ["instance_method"])
    @pytest.mark.parametrize(
        "callback",
        [
            call_history.instance_callback,
            CallHistory.class_callback,
            CallHistory.static_callback,
        ],
    )
    @pytest.mark.parametrize("args,kwargs", [([], {}), ([1, 2.0], {"kwarg1": 0, "kwarg2": None})])
    def test_register_method_callback(self, event_name, callback, args, kwargs):
        """Methods' callbacks are invoked in the correct order and with the correct arguments."""
        callbackable = TestCallbackable.MyCallbackable()

        # No callback is called before registration.
        event_sequence = [event_name]
        self.check_method_call_order(callbackable, event_name, event_sequence, *args, **kwargs)

        # Register the callback.
        callbackable.register_callback(event_name, callback)

        # After the registration, the callback is invoked correctly.
        event_sequence = [event_name, "callback_" + event_name]
        self.check_method_call_order(callbackable, event_name, event_sequence, *args, **kwargs)

    def test_register_magic_method_callback(self):
        """Callbacks registered to magic methods are invoked correctly."""
        callbackable = TestCallbackable.MyCallbackable()
        callbackable.register_callback("__iadd__", call_history.instance_callback)

        extension = [1, 2]
        call_history.reset_history()
        callbackable += extension
        assert call_history.history == [
            ["__iadd__", (extension,)],
            ["callback___iadd__", (extension,)],
        ]

    def test_register_event_callback(self):
        """Callbacks registered to a event are handled corectly."""
        callbackable = TestCallbackable.MyCallbackable()

        # Register the callbacks to event1 (event_method1 and event_method2).
        callbackable.register_callback("event1", call_history.instance_callback)
        callbackable.register_callback("event1", CallHistory.class_callback)
        # Register one callback to event2 (only event_method2).
        callbackable.register_callback("event2", CallHistory.static_callback)

        # Check the event sequence for both methods belong to the two events.
        event_sequence = [
            "event_method1",
            "callback_event_method1",
            "callback_event_method1",
        ]
        self.check_method_call_order(callbackable, "event_method1", event_sequence)

        event_sequence = [
            "event_method2",
            "callback_event_method2",
            "callback_event_method2",
            "callback_event_method2",
        ]
        self.check_method_call_order(callbackable, "event_method2", event_sequence)

    def test_not_callback_method_raise_exception(self):
        """An exception is raised if a callback is registered for a method not tagged with callback_method."""

        class TempCallbackable(Callbackable):
            def not_callback_method(self):
                pass

        callbackable = TempCallbackable()
        with pytest.raises(
            CallbackRegistrationError,
            match="is not tagged with the @callback_method decorator",
        ):
            callbackable.register_callback("not_callback_method", call_history.instance_callback)

    def test_unknown_event_raise_exception(self):
        """An exception is raised if a callback is registered for an unknown callback event."""
        callbackable = TestCallbackable.MyCallbackable()
        with pytest.raises(
            CallbackRegistrationError,
            match='is associated to the callback event "unknown"',
        ):
            callbackable.register_callback("unknown", call_history.instance_callback)



"""
Test classes and function in module openff.toolkit.typing.engines.smirnoff.plugins
"""

import pytest

from openff.toolkit import ForceField, Quantity
from openff.toolkit.typing.engines.smirnoff.plugins import (
    _load_handler_plugins,
    load_handler_plugins,
)


def test_force_field_custom_handler():
    """Tests a force field can make use of a custom parameter handler registered
    through the entrypoint plugin system.
    """

    # Construct a simple FF which only uses the custom handler.
    force_field_contents = "\n".join(
        [
            "<?xml version='1.0' encoding='ASCII'?>",
            "<SMIRNOFF version='0.3' aromaticity_model='OEAroModel_MDL'>",
            "  <CustomHandler version='0.3'></CustomHandler>",
            "</SMIRNOFF>",
        ]
    )

    # An exception should be raised when plugins aren't allowed.
    with pytest.raises(KeyError) as error_info:
        ForceField(force_field_contents)

    assert "Cannot find a registered parameter handler class for tag 'CustomHandler'" in error_info.value.args[0]

    # Otherwise the FF should be created as expected.
    force_field = ForceField(force_field_contents, load_plugins=True)

    parameter_handler = force_field.get_parameter_handler("CustomHandler")
    assert parameter_handler is not None
    assert parameter_handler.__class__.__name__ == "CustomHandler"

    assert parameter_handler.__class__ in force_field._plugin_parameter_handler_classes


def test_load_handler_plugins():
    """Tests that parameter handlers can be registered as plugins."""

    registered_plugins = load_handler_plugins()

    assert len(registered_plugins) == 2

    registered_plugin_names = [plugin.__name__ for plugin in registered_plugins]

    assert "CustomHandler" in registered_plugin_names
    assert "FOOBuckinghamHandler" in registered_plugin_names


def test_do_not_load_other_type():
    with pytest.raises(NotImplementedError, match="foobar"):
        _load_handler_plugins(handler_name="foobar", expected_type=type(None))


@pytest.mark.skip(reason="Broken around 8/29/23, unclear why")
def test_skip_wrong_subclass(caplog):
    import logging

    caplog.set_level(logging.INFO)
    load_handler_plugins()

    assert "does not inherit from ParameterHandler" in caplog.text, caplog.text


def test_buckingham_type():
    """Reproduce, in part, issue #1888."""
    from custom_plugins.handler_plugins import FOOBuckinghamHandler

    parameter = FOOBuckinghamHandler.FOOBuckinghamType(
        smirks="[*:1]",
        a="2 kilojoule_per_mole",
        b="1/nanometer",
        c="-0.5 kilojoule_per_mole * nanometer**6",
    )

    for param in ["a", "b", "c"]:
        assert isinstance(getattr(parameter, param), Quantity)

    assert str(parameter.a.units) == "kilojoule_per_mole"
    assert str(parameter.b.units) == "1 / nanometer"
    assert str(parameter.c.units) == "kilojoule_per_mole * nanometer ** 6"



"""
Base class for toolkit wrappers. Defines the public API and some shared methods
"""

__all__ = ("ToolkitWrapper",)

from functools import wraps
from typing import TYPE_CHECKING, TypedDict

from openff.toolkit.utils.constants import DEFAULT_AROMATICITY_MODEL
from openff.toolkit.utils.exceptions import (
    IncorrectNumConformersError,
    IncorrectNumConformersWarning,
    ToolkitUnavailableException,
)

if TYPE_CHECKING:
    from openff.toolkit.topology.molecule import Molecule


class _ChargeSettings(TypedDict, total=False):
    n_conformers: int
    rec_confs: int
    min_confs: int
    max_confs: int  # OpenEyeToolkitWrapper sometimes defines this as None
    antechamber_keyword: str
    oe_charge_method: str


def _mol_to_ctab_and_aro_key(self, molecule: "Molecule", aromaticity_model=DEFAULT_AROMATICITY_MODEL) -> str:
    return f"{molecule.ordered_connection_table_hash()}-{aromaticity_model}"


class ToolkitWrapper:
    """
    Base class for wrappers around external software toolkits.

    .. warning :: This API is experimental and subject to change.

    See also
    ========
    ToolkitRegistry, OpenEyeToolkitWrapper, RDKitToolkitWrapper,
    AmberToolsToolkitWrapper, NAGLToolkitWrapper, BuiltInToolkitWrapper

    """

    _is_available: bool | None = None  # True if toolkit is available
    _toolkit_version: str | None = None
    _toolkit_name: str | None = None  # Name of the toolkit
    _toolkit_installation_instructions: str | None = None  # Installation instructions for the toolkit
    _supported_charge_methods: dict[str, _ChargeSettings] = dict()
    _toolkit_file_read_formats: list[str] = list()
    _toolkit_file_write_formats: list[str] = list()

    # @staticmethod
    # TODO: Right now, to access the class definition, I have to make this a classmethod
    # and thereby call it with () on the outermost decorator. Is this wasting time? Are we caching
    # the is_available results?
    @classmethod
    def requires_toolkit(cls):  # remember cls is a ToolkitWrapper subclass here
        def decorator(func):
            @wraps(func)
            def wrapped_function(*args, **kwargs):
                if not cls.is_available():
                    msg = f"This function requires the {cls._toolkit_name} toolkit"
                    raise ToolkitUnavailableException(msg)
                value = func(*args, **kwargs)
                return value

            return wrapped_function

        return decorator

    @property
    def toolkit_name(self):
        """
        Return the name of the toolkit wrapped by this class as a str

        .. warning :: This API is experimental and subject to change.

        Returns
        -------
        toolkit_name
            The name of the wrapped toolkit

        """
        return self.__class__._toolkit_name

    @property
    def toolkit_installation_instructions(self):
        """
        Instructions on how to install the wrapped toolkit.
        """
        return self._toolkit_installation_instructions

    @property
    def toolkit_file_read_formats(self):
        """
        List of file formats that this toolkit can read.
        """
        return self._toolkit_file_read_formats

    @property
    def toolkit_file_write_formats(self) -> list[str]:
        """
        List of file formats that this toolkit can write.
        """
        return self._toolkit_file_write_formats

    @property
    def supported_charge_methods(self) -> list[str]:
        return [*self._supported_charge_methods.keys()]

    @classmethod
    def is_available(cls):
        """
        Check whether the corresponding toolkit can be imported

        Returns
        -------
        is_installed
            True if corresponding toolkit is installed, False otherwise.

        """
        return NotImplementedError

    @property
    def toolkit_version(self):
        """
        Return the version of the wrapped toolkit as a str

        .. warning :: This API is experimental and subject to change.

        Returns
        -------
        toolkit_version
            The version of the wrapped toolkit

        """
        return self._toolkit_version

    def from_file(self, file_path, file_format, allow_undefined_stereo=False):
        """
        Return an openff.toolkit.topology.Molecule from a file using this toolkit.

        Parameters
        ----------
        file_path
            The file to read the molecule from
        file_format
            Format specifier, usually file suffix (eg. 'MOL2', 'SMI')
            Note that not all toolkits support all formats. Check
            ToolkitWrapper.toolkit_file_read_formats for details.
        allow_undefined_stereo
            If false, raises an exception if any molecules contain undefined stereochemistry.
        _cls
            Molecule constructor
        Returns
        -------
        molecules
            a list of Molecule objects is returned.

        """
        return NotImplementedError

    def from_file_obj(self, file_obj, file_format, allow_undefined_stereo=False, _cls=None):
        """
        Return an openff.toolkit.topology.Molecule from a file-like object (an object with
        a ".read()" method using this toolkit.


        Parameters
        ----------
        file_obj
            The file-like object to read the molecule from
        file_format
            Format specifier, usually file suffix (eg. 'MOL2', 'SMI')
            Note that not all toolkits support all formats. Check
            ToolkitWrapper.toolkit_file_read_formats for details.
        allow_undefined_stereo
            If false, raises an exception if any molecules contain undefined stereochemistry.
            If false, the function skips loading the molecule.
        _cls
            Molecule constructor

        Returns
        -------
        molecules
            a list of Molecule objects is returned.
        """
        return NotImplementedError

    def _check_n_conformers(
        self,
        molecule: "Molecule",
        partial_charge_method: str | None = None,
        min_confs: int | None = None,
        max_confs: int | None = None,
        strict_n_conformers: bool = False,
    ):
        """
        Private method for validating the number of conformers on a molecule prior to partial
        charge calculation

        Parameters
        ----------
        molecule
            Molecule for which partial charges are to be computed
        partial_charge_method
            The name of the charge method being used
        min_confs
            The minimum number of conformers required to use this charge method
        max_confs
            The maximum number of conformers required to use this charge method
        strict_n_conformers
            Whether to raise an exception if an invalid number of conformers is provided.
            If this is False and an invalid number of conformers is found, a warning will be raised.

        Raises
        ------
        IncorrectNumConformersError
            If the wrong number of conformers is attached to the input molecule, and
            strict_n_conformers is True.
        """
        import warnings

        n_confs = molecule.n_conformers
        wrong_confs_msg = (
            f"Molecule '{molecule}' has {n_confs} conformers, but charge method '{partial_charge_method}' expects"
        )
        exception_suffix = (
            "You can disable this error by setting `strict_n_conformers=False' "
            "when calling 'molecule.assign_partial_charges'."
        )
        # If there's no n_confs filter, then this molecule automatically passes
        if min_confs is None and max_confs is None:
            return
        # If there's constraints on both ends, check both limits
        elif min_confs is not None and max_confs is not None:
            if not (min_confs <= n_confs <= max_confs):
                if min_confs == max_confs:
                    wrong_confs_msg += f" exactly {min_confs}."
                else:
                    wrong_confs_msg += f" between {min_confs} and {max_confs}."

            else:
                return
        # If there's only a max constraint, check that
        elif min_confs is not None and max_confs is None:
            if not (min_confs <= n_confs):
                wrong_confs_msg += f" at least {min_confs}."
            else:
                return
        # If there's only a maximum constraint, check that
        elif min_confs is None and max_confs is not None:
            if not (n_confs <= max_confs):
                wrong_confs_msg += f" at most {max_confs}."
            else:
                return
        # If we've made it this far, the molecule has the wrong number of conformers
        if strict_n_conformers:
            wrong_confs_msg += exception_suffix
            raise IncorrectNumConformersError(wrong_confs_msg)
        else:
            warnings.warn(wrong_confs_msg, IncorrectNumConformersWarning, stacklevel=2)

    def __repr__(self):
        return f"ToolkitWrapper around {self.toolkit_name} version {self.toolkit_version}"



__all__ = (
    "ALLOWED_AROMATICITY_MODELS",
    "ALLOWED_CHARGE_MODELS",
    "ALLOWED_FRACTIONAL_BOND_ORDER_MODELS",
    "DEFAULT_AROMATICITY_MODEL",
    "DEFAULT_CHARGE_MODEL",
    "DEFAULT_FRACTIONAL_BOND_ORDER_MODEL",
)


ALLOWED_AROMATICITY_MODELS = ["OEAroModel_MDL"]
DEFAULT_AROMATICITY_MODEL = ALLOWED_AROMATICITY_MODELS[0]

ALLOWED_FRACTIONAL_BOND_ORDER_MODELS = ["Wiberg"]
DEFAULT_FRACTIONAL_BOND_ORDER_MODEL = ALLOWED_FRACTIONAL_BOND_ORDER_MODELS[0]

ALLOWED_CHARGE_MODELS = ["AM1-BCC"]
DEFAULT_CHARGE_MODEL = ALLOWED_CHARGE_MODELS[0]



import json
import sys

release_tag = sys.argv[1]
python_version = sys.argv[2]
ci_os = sys.argv[3]

platform_mapping = {
    "ubuntu-latest": "linux-64",
    "macOS-latest": "osx-64",
}

data = {
    "name": "openff-toolkit",
    "channel": "conda-forge",
    "python": [python_version],
    "platform": [platform_mapping[ci_os]],
    "release": release_tag,
}

with open("new_cookiecutter.json", "w") as fp:
    json.dump(data, fp)



from openff.toolkit import unit
from openff.toolkit.typing.engines.smirnoff import ParameterHandler, ParameterIOHandler
from openff.toolkit.typing.engines.smirnoff.parameters import (
    ParameterAttribute,
    ParameterType,
    _NonbondedHandler,
)


class CustomHandler(ParameterHandler):
    _TAGNAME = "CustomHandler"


class WrongSubclass(list):
    _TAGNAME = "CustomHandler"


class CustomIOHandler(ParameterIOHandler):
    _FORMAT = "JSON"


class FOOBuckinghamHandler(_NonbondedHandler):
    """A custom parameter handler for buckingham interactions."""

    class FOOBuckinghamType(ParameterType):
        """A custom parameter type for buckingham interactions."""

        _ELEMENT_NAME = "Atom"

        # Define unit as a Unit object
        a = ParameterAttribute(
            default=None,
            unit=unit.kilojoule_per_mole,
        )

        # Define using a string
        b = ParameterAttribute(
            default=None,
            unit="nanometer**-1",
        )

        c = ParameterAttribute(
            default=None,
            unit=unit.kilojoule_per_mole * unit.nanometer**6,
        )

    _TAGNAME = "FOOBuckingham"
    _INFOTYPE = FOOBuckinghamType



"""
Wrapper class providing a minimal consistent interface to `AmberTools <http://ambermd.org/AmberTools.php>`_.
"""

__all__ = ("AmberToolsToolkitWrapper",)

import functools
import subprocess
import tempfile
from collections import defaultdict
from shutil import which
from typing import TYPE_CHECKING

import numpy as np

from openff.toolkit import Quantity, unit
from openff.toolkit.utils import base_wrapper, rdkit_wrapper
from openff.toolkit.utils.exceptions import (
    AntechamberNotFoundError,
    ChargeCalculationError,
    ChargeMethodUnavailableError,
    ToolkitUnavailableException,
)

if TYPE_CHECKING:
    from openff.toolkit.topology.molecule import Molecule


class AmberToolsToolkitWrapper(base_wrapper.ToolkitWrapper):
    """
    AmberTools toolkit wrapper

    .. warning :: This API is experimental and subject to change.
    """

    _toolkit_name = "AmberTools"
    _toolkit_installation_instructions = (
        "The AmberTools toolkit (free and open source) can be found at https://anaconda.org/conda-forge/ambertools"
    )
    _supported_charge_methods = {
        "am1bcc": {
            "antechamber_keyword": "bcc",
            "min_confs": 1,
            "max_confs": 1,
            "rec_confs": 1,
        },
        "am1-mulliken": {
            "antechamber_keyword": "mul",
            "min_confs": 1,
            "max_confs": 1,
            "rec_confs": 1,
        },
        "gasteiger": {
            "antechamber_keyword": "gas",
            "min_confs": 0,
            "max_confs": 0,
            "rec_confs": 0,
        },
    }

    SUPPORTED_CHARGE_METHODS = _supported_charge_methods

    def __init__(self):
        super().__init__()

        if not self.is_available():
            raise ToolkitUnavailableException(
                f"The required toolkit {self._toolkit_name} is not "
                f"available. {self._toolkit_installation_instructions}"
            )

        # TODO: Find AMBERHOME or executable home, checking miniconda if needed
        # Store an instance of an RDKitToolkitWrapper for file I/O
        self._rdkit_toolkit_wrapper = rdkit_wrapper.RDKitToolkitWrapper()

    @functools.cached_property
    def _toolkit_version(self):
        from openff.utilities.provenance import get_ambertools_version

        return get_ambertools_version()

    @staticmethod
    def is_available() -> bool:
        """
        Check whether the AmberTools toolkit is installed

        Returns
        -------
        is_installed
            True if AmberTools is installed, False otherwise.

        """
        # TODO: Check all tools needed
        ANTECHAMBER_PATH = which("antechamber")
        if ANTECHAMBER_PATH is None:
            return False
        # AmberToolsToolkitWrapper needs RDKit to do basically anything, since its interface requires SDF I/O
        if not (rdkit_wrapper.RDKitToolkitWrapper.is_available()):
            return False
        return True

    def assign_partial_charges(
        self,
        molecule: "Molecule",
        partial_charge_method: str | None = None,
        use_conformers: list[Quantity] | None = None,
        strict_n_conformers: bool = False,
        normalize_partial_charges: bool = True,
        _cls=None,
    ):
        """
        Compute partial charges with AmberTools using antechamber/sqm, and assign
        the new values to the partial_charges attribute.

        .. warning :: This API experimental and subject to change.

        .. todo ::

           * Do we want to also allow ESP/RESP charges?

        Parameters
        ----------
        molecule
            Molecule for which partial charges are to be computed
        partial_charge_method
            The charge model to use. One of ['gasteiger', 'am1bcc', 'am1-mulliken'].
            If None, 'am1-mulliken' will be used.
        use_conformers
            with shape (n_atoms, 3) and dimension of distance. Optional, default = None
            List of unit-wrapped numpy arrays to use for partial charge calculation.
            If None, an appropriate number of conformers will be generated.
        strict_n_conformers
            Whether to raise an exception if an invalid number of conformers is provided for
            the given charge method.
            If this is False and an invalid number of conformers is found, a warning will be raised.
        normalize_partial_charges
            Whether to offset partial charges so that they sum to the total formal charge of the molecule.
            This is used to prevent accumulation of rounding errors when the partial charge generation method has
            low precision.
        _cls
            Molecule constructor

        Raises
        ------
        ChargeMethodUnavailableError if the requested charge method can not be handled by this toolkit

        ChargeCalculationError if the charge method is supported by this toolkit, but fails
        """

        import os
        import subprocess

        from openff.toolkit.topology import Molecule

        if partial_charge_method is None:
            partial_charge_method = "am1-mulliken"
        else:
            # Standardize method name for string comparisons
            partial_charge_method = partial_charge_method.lower()

        if partial_charge_method not in self._supported_charge_methods:
            raise ChargeMethodUnavailableError(
                f"partial_charge_method '{partial_charge_method}' is not available from AmberToolsToolkitWrapper. "
                f"Available charge methods are {self._supported_charge_methods}"
            )

        charge_method = self._supported_charge_methods[partial_charge_method]

        if _cls is None:
            _cls = Molecule

        # Make a temporary copy of the molecule, since we'll be messing with its conformers
        mol_copy = _cls(molecule)

        if use_conformers is None:
            if charge_method["rec_confs"] == 0:
                mol_copy._conformers = None
            else:
                mol_copy.generate_conformers(
                    n_conformers=charge_method["rec_confs"],
                    rms_cutoff=0.25 * unit.angstrom,
                    toolkit_registry=rdkit_wrapper.RDKitToolkitWrapper(),
                )
            # TODO: What's a "best practice" RMS cutoff to use here?
        else:
            mol_copy._conformers = None
            for conformer in use_conformers:
                mol_copy._add_conformer(conformer)
            self._check_n_conformers(
                mol_copy,
                partial_charge_method=partial_charge_method,
                min_confs=charge_method["min_confs"],
                max_confs=charge_method["max_confs"],
                strict_n_conformers=strict_n_conformers,
            )

        ANTECHAMBER_PATH = which("antechamber")
        if ANTECHAMBER_PATH is None:
            raise AntechamberNotFoundError("Antechamber not found, cannot run assign_partial_charges()")

        # Compute charges
        with tempfile.TemporaryDirectory() as tmpdir:
            net_charge = mol_copy.total_charge.m_as(unit.elementary_charge)
            # Write out molecule in SDF format
            # TODO: How should we handle multiple conformers?
            self._rdkit_toolkit_wrapper.to_file(mol_copy, f"{tmpdir}/molecule.sdf", file_format="sdf")
            # Compute desired charges
            # TODO: Add error handling if antechamber chokes
            short_charge_method = charge_method["antechamber_keyword"]
            subprocess.check_output(
                [
                    "antechamber",
                    "-i",
                    "molecule.sdf",
                    "-fi",
                    "sdf",
                    "-o",
                    "charged.mol2",
                    "-fo",
                    "mol2",
                    "-pf",
                    "yes",
                    "-dr",
                    "n",
                    "-c",
                    str(short_charge_method),
                    "-nc",
                    str(net_charge),
                ],
                cwd=tmpdir,
            )
            # Write out just charges
            subprocess.check_output(
                [
                    "antechamber",
                    "-dr",
                    "n",
                    "-i",
                    "charged.mol2",
                    "-fi",
                    "mol2",
                    "-o",
                    "charges2.mol2",
                    "-fo",
                    "mol2",
                    "-c",
                    "wc",
                    "-cf",
                    "charges.txt",
                    "-pf",
                    "yes",
                ],
                cwd=tmpdir,
            )
            # Check to ensure charges were actually produced
            if not os.path.exists(f"{tmpdir}/charges.txt"):
                # TODO: copy files into local directory to aid debugging?
                raise ChargeCalculationError(
                    "Antechamber/sqm partial charge calculation failed on "
                    f"molecule {molecule.name} (SMILES {molecule.to_smiles()})"
                )
            # Read the charges
            with open(f"{tmpdir}/charges.txt") as infile:
                contents = infile.read()
            text_charges = contents.split()
            charges_array = np.zeros([molecule.n_atoms], np.float64)
            for index, token in enumerate(text_charges):
                charges_array[index] = float(token)
            # TODO: Ensure that the atoms in charged.mol2 are in the same order as in molecule.sdf
        molecule.partial_charges = Quantity(charges_array, unit.elementary_charge)

        if normalize_partial_charges:
            molecule._normalize_partial_charges()

    def _modify_sqm_in_to_request_bond_orders(self, file_path):
        """
        Modify a sqm.in file produced by antechamber to include the "printbondorders=1" directive
        in the header. This method will overwrite the original file.

        Parameters
        ----------
        file_path
            The path to sqm.in
        """

        data = open(file_path).read()

        # Original sqm.in file headerlooks like:

        # Run semi-empirical minimization
        #  &qmmm
        #    qm_theory='AM1', grms_tol=0.0005,
        #    scfconv=1.d-10, ndiis_attempts=700,   qmcharge=0,
        #  /
        # ... (atom coordinates in something like XYZ format) ...

        # To get WBOs, we need to add "printbondorders=1" to the list of keywords

        # First, split the sqm.in text at the "/" mark at the end of the header
        datasp = data.split("/")
        # Insert the "printbondorders" directive in a new line and re-add the "/"
        datasp.insert(1, "printbondorders=1, \n /")
        # Reassemble the file text
        new_data = "".join(datasp)
        # Write the new file contents, overwriting the original file.
        with open(file_path, "w") as of:
            of.write(new_data)

    def _get_fractional_bond_orders_from_sqm_out(self, file_path, validate_elements=None):
        """
        Process a SQM output file containing bond orders, and return a dict of the form
        dict[atom_1_index, atom_2_index] = fractional_bond_order

        Parameters
        ----------
        file_path
            File path for sqm output file
        validate_elements
            The element symbols expected in molecule index order. A ValueError will be raised
            if the elements are not found in this order.

        Returns
        -------
        bond_orders
            A dictionary where the keys are tuples of two atom indices and the values are
            floating-point bond orders. The keys are sorted in ascending order, such that
            the lower atom index is key[0] and the higher is key[1].
        """

        # Example sqm.out section with WBOs:
        #  Bond Orders
        #
        #   QMMM:    NUM1 ELEM1 NUM2 ELEM2      BOND_ORDER
        #   QMMM:       2   C      1   C        1.41107532
        #   QMMM:       3   C      1   C        1.41047804
        # ...
        #   QMMM:      15   H     13   H        0.00000954
        #   QMMM:      15   H     14   H        0.00000813
        #
        #            --------- Calculation Completed ----------

        data = open(file_path).read()

        begin_sep = """ Bond Orders
 
  QMMM:    NUM1 ELEM1 NUM2 ELEM2      BOND_ORDER
"""  # noqa
        end_sep = """

           --------- Calculation Completed ----------
"""
        # Extract the chunk of text between begin_sep and end_sep, and split it by newline
        fbo_lines = data.split(begin_sep)[1].split(end_sep)[0].split("\n")

        # Iterate over the lines and populate the dict to return
        bond_orders = dict()
        for line in fbo_lines:
            linesp = line.split()
            atom_index_1 = int(linesp[1])
            atom_element_1 = linesp[2]
            atom_index_2 = int(linesp[3])
            atom_element_2 = linesp[4]
            bond_order = float(linesp[5])

            # If validate_elements was provided, ensure that the ordering of element symbols is what we expected
            if validate_elements is not None:
                if (atom_element_1 != validate_elements[atom_index_1 - 1]) or (
                    atom_element_2 != validate_elements[atom_index_2 - 1]
                ):
                    # raise ValueError('\n'.join(fbo_lines))
                    raise ValueError(
                        f"Elements or indexing in sqm output differ from expectation. "
                        f"Expected {validate_elements[atom_index_1]} with index {atom_index_1} and "
                        f"{validate_elements[atom_index_2]} with index {atom_index_2}, "
                        f"but SQM output has {atom_element_1} and {atom_element_2} for the same atoms."
                    )

            # To make lookup easier, we identify bonds as integer tuples with the lowest atom index
            # first and the highest second.
            index_tuple = tuple(sorted([atom_index_1, atom_index_2]))
            bond_orders[index_tuple] = bond_order
        return bond_orders

    def assign_fractional_bond_orders(
        self,
        molecule: "Molecule",
        bond_order_model: str | None = None,
        use_conformers: list[str] | None = None,
        _cls=None,
    ):
        """
        Update and store list of bond orders this molecule. Bond orders are stored on each
        bond, in the `bond.fractional_bond_order` attribute.

        .. warning :: This API is experimental and subject to change.

        Parameters
        ----------
        molecule
            The molecule to assign wiberg bond orders to
        bond_order_model
            The charge model to use. Only allowed value is 'am1-wiberg'. If None, 'am1-wiberg' will be used.
        use_conformers
            The conformers to use for fractional bond order calculation. If None, an appropriate
            number of conformers will be generated by an available ToolkitWrapper.
        _cls
            Molecule constructor
        """
        from openff.toolkit.topology import Molecule

        ANTECHAMBER_PATH = which("antechamber")
        if ANTECHAMBER_PATH is None:
            raise AntechamberNotFoundError(
                "Antechamber not found, cannot run AmberToolsToolkitWrapper.assign_fractional_bond_orders()"
            )

        if _cls is None:
            _cls = Molecule

        # Make a copy since we'll be messing with this molecule's conformers
        temp_mol = _cls(molecule)

        if use_conformers is None:
            temp_mol.generate_conformers(
                n_conformers=1,
                toolkit_registry=self._rdkit_toolkit_wrapper,
            )
        else:
            temp_mol._conformers = None
            for conformer in use_conformers:
                temp_mol._add_conformer(conformer)

        if len(temp_mol.conformers) == 0:
            raise ValueError(
                "No conformers present in molecule submitted for fractional bond order calculation. Consider "
                "loading the molecule from a file with geometry already present or running "
                "molecule.generate_conformers() before calling molecule.assign_fractional_bond_orders"
            )

        # Compute bond orders
        bond_order_model_to_antechamber_keyword = {"am1-wiberg": "mul"}
        supported_bond_order_models = list(bond_order_model_to_antechamber_keyword.keys())
        if bond_order_model is None:
            bond_order_model = "am1-wiberg"

        bond_order_model = bond_order_model.lower()

        if bond_order_model not in supported_bond_order_models:
            raise ValueError(
                f"Bond order model '{bond_order_model}' is not supported by AmberToolsToolkitWrapper. "
                f"Supported models are {supported_bond_order_models}"
            )
        ac_charge_keyword = bond_order_model_to_antechamber_keyword[bond_order_model]

        bond_orders = defaultdict(list)

        for conformer in [*temp_mol.conformers]:
            with tempfile.TemporaryDirectory() as tmpdir:
                net_charge = temp_mol.total_charge
                # Write out molecule in SDF format
                temp_mol._conformers = [conformer]
                self._rdkit_toolkit_wrapper.to_file(temp_mol, f"{tmpdir}/molecule.sdf", file_format="sdf")
                # Prepare sqm.in file as if we were going to run charge calc
                # TODO: Add error handling if antechamber chokes
                subprocess.check_output(
                    [
                        "antechamber",
                        "-i",
                        "molecule.sdf",
                        "-fi",
                        "sdf",
                        "-o",
                        "sqm.in",
                        "-fo",
                        "sqmcrt",
                        "-pf",
                        "yes",
                        "-c",
                        ac_charge_keyword,
                        "-nc",
                        str(net_charge),
                    ],
                    cwd=tmpdir,
                )
                # Modify sqm.in to request bond order calculation
                self._modify_sqm_in_to_request_bond_orders(f"{tmpdir}/sqm.in")
                # Run sqm to get bond orders
                subprocess.check_output(["sqm", "-i", "sqm.in", "-o", "sqm.out", "-O"], cwd=tmpdir)
                # Ensure that antechamber/sqm did not change the indexing by checking against
                # an ordered list of element symbols for this molecule
                expected_elements = [atom.symbol for atom in molecule.atoms]
                conformer_bond_orders = self._get_fractional_bond_orders_from_sqm_out(
                    f"{tmpdir}/sqm.out", validate_elements=expected_elements
                )

                for bond_indices, value in conformer_bond_orders.items():
                    bond_orders[bond_indices].append(value)

        # Note that sqm calculate WBOs for ALL PAIRS of atoms, not just those that have
        # bonds defined in the original molecule. So here we iterate over the bonds in
        # the original molecule and only nab the WBOs for those.
        for bond in molecule.bonds:
            # The atom index tuples that act as bond indices are ordered from lowest to highest by
            # _get_fractional_bond_orders_from_sqm_out, so here we make sure that we look them up in
            # sorted order as well
            sorted_atom_indices = sorted(tuple([bond.atom1_index + 1, bond.atom2_index + 1]))
            bond.fractional_bond_order = np.mean(bond_orders[tuple(sorted_atom_indices)])



import mdtraj.utils
import os
import shutil
import logging
logger = logging.getLogger(__name__)
from openmoltools.utils import getoutput

# Functionality below adapted from openmoltools.amber.run_tleap

def run_tleap(molecule_name, mol2_filename, frcmod_filename, prmtop_filename=None, inpcrd_filename=None, log_debug_output=False, verbose = True):
    """Run AmberTools tleap to create simulation files for AMBER
    Parameters
    ----------
    molecule_name : str
        The name of the molecule
    mol2_filename : str
        mol2 filename with target atom type names
    frcmod_filename : str
        Amber frcmod file produced by prmchk
    prmtop_filename : str, optional, default=None
        Amber prmtop file produced by tleap, defaults to molecule_name
    inpcrd_filename : str, optional, default=None
        Amber inpcrd file produced by tleap, defaults to molecule_name
    log_debug_output : bool, optional, default=False
        If true, will send output of tleap to logger.
    verbose : bool
        Be verbose
    Returns
    -------
    prmtop_filename : str
        Amber prmtop file produced by tleap
    inpcrd_filename : str
        Amber inpcrd file produced by tleap
    """
    if prmtop_filename is None:
        prmtop_filename = f"{molecule_name}.prmtop"
    if inpcrd_filename is None:
        inpcrd_filename = f"{molecule_name}.inpcrd"

    #Get absolute paths for input/output
    mol2_filename = os.path.abspath( mol2_filename )
    frcmod_filename = os.path.abspath( frcmod_filename )
    prmtop_filename = os.path.abspath( prmtop_filename )
    inpcrd_filename = os.path.abspath( inpcrd_filename )

    #Work in a temporary directory, on hard coded filenames, to avoid any issues AMBER may have with spaces and other special characters in filenames
    with mdtraj.utils.enter_temp_directory():
        shutil.copy( mol2_filename, 'file.mol2' )
        shutil.copy( frcmod_filename, 'file.frcmod' )

        tleap_input = """
    source oldff/leaprc.ff99
    LIG = loadmol2 file.mol2
    loadamberparams file.frcmod
    check LIG
    saveamberparm LIG out.prmtop out.inpcrd
    quit
"""

        file_handle = open('tleap_commands', 'w')
        file_handle.writelines(tleap_input)
        file_handle.close()

        cmd = f"tleap -f {file_handle.name} "
        if log_debug_output: logger.debug(cmd)
        if verbose: print(cmd)

        output = getoutput(cmd)
        if log_debug_output: logger.debug(output)
        if verbose: print(output)

        check_for_errors( output, other_errors = ['Improper number of arguments'], ignore_errors=['Perhaps a format error'])

        #Copy back target files
        shutil.copy( 'out.prmtop', prmtop_filename )
        shutil.copy( 'out.inpcrd', inpcrd_filename )

    return prmtop_filename, inpcrd_filename


def check_for_errors( outputtext, other_errors = None, ignore_errors = None ):
    """Check AMBER package output for the string 'ERROR' (upper or lowercase) and (optionally) specified other strings and raise an exception if it is found (to avoid silent failures which might be noted to log but otherwise ignored).
    Parameters
    ----------
    outputtext : str
        String listing output text from an (AMBER) command which should be checked for errors.
    other_errors : list(str), default None
        If specified, provide strings for other errors which will be chcked for, such as "improper number of arguments", etc.
    ignore_errors: list(str), default None
        If specified, AMBER output lines containing errors but also containing any of the specified strings will be ignored (because, for example, AMBER issues an "ERROR" for non-integer charges in some cases when only a warning is needed).
    Notes
    -----
    If error(s) are found, raise a RuntimeError and attept to print the appropriate errors from the processed text."""
    lines = outputtext.split('\n')
    error_lines = []
    for line in lines:
        if 'ERROR' in line.upper():
            error_lines.append( line )
        if not other_errors == None:
            for err in other_errors:
                if err.upper() in line.upper():
                    error_lines.append( line )

    if not ignore_errors == None and len(error_lines)>0:
        new_error_lines = []
        for ign in ignore_errors:
            ignore = False
            for err in error_lines:
                if ign in err:
                    ignore = True
            if not ignore:
                new_error_lines.append( err )
        error_lines = new_error_lines

    if len(error_lines) > 0:
        print("Unexpected errors encountered running AMBER tool. Offending output:")
        for line in error_lines: print(line)
        raise(RuntimeError("Error encountered running AMBER tool. Exiting."))

    return



from openff.toolkit.topology.molecule import (
    Atom,
    Bond,
    FrozenMolecule,
    HierarchyElement,
    HierarchyScheme,
    Molecule,
    Particle,
)
from openff.toolkit.topology.topology import (
    DuplicateUniqueMoleculeError,
    ImproperDict,
    InvalidBoxVectorsError,
    InvalidPeriodicityError,
    MissingUniqueMoleculesError,
    NotBondedError,
    SortedDict,
    TagSortedDict,
    Topology,
    UnsortedDict,
    ValenceDict,
)

__all__ = [
    "Atom",
    "Bond",
    "DuplicateUniqueMoleculeError",
    "FrozenMolecule",
    "HierarchyElement",
    "HierarchyScheme",
    "ImproperDict",
    "InvalidBoxVectorsError",
    "InvalidPeriodicityError",
    "MissingUniqueMoleculesError",
    "Molecule",
    "NotBondedError",
    "Particle",
    "SortedDict",
    "TagSortedDict",
    "Topology",
    "UnsortedDict",
    "ValenceDict",
]



#!/usr/bin/env python3
#
# openff-toolkit documentation build configuration file, created by
# sphinx-quickstart on Sun Dec  3 23:12:54 2017.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys
from importlib.util import find_spec as find_import_spec

import openff.toolkit

sys.path.insert(0, os.path.abspath("."))


# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = "4.4.0"

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "sphinx.ext.autosummary",
    "sphinx.ext.doctest",
    "sphinx.ext.todo",
    "sphinx.ext.mathjax",
    "sphinx.ext.viewcode",
    "sphinx.ext.intersphinx",
    "openff_sphinx_theme",
    "myst_nb",
    # "myst_parser",
]

# Autodoc settings
autosummary_generate = True

autodoc_default_options = {
    "members": True,
    "inherited-members": True,
    "member-order": "bysource",
}
autodoc_preserve_defaults = True
autodoc_typehints_format = "short"
# Workaround for autodoc_typehints_format not working for attributes
# see https://github.com/sphinx-doc/sphinx/issues/10290#issuecomment-1079740009
python_use_unqualified_type_names = True

napoleon_numpy_docstring = True
napoleon_google_docstring = False
napoleon_attr_annotations = True
napoleon_custom_sections = [("attributes", "params_style")]
napoleon_use_rtype = False
napoleon_use_param = True

_python_doc_base = "https://docs.python.org/3.7"
intersphinx_mapping = {
    "python": ("https://docs.python.org/3.7", None),
    "numpy": ("https://numpy.org/doc/stable", None),
    "scipy": ("https://docs.scipy.org/doc/scipy", None),
    "scikit.learn": ("https://scikit-learn.org/stable", None),
    "openmm": ("http://docs.openmm.org/latest/api-python/", None),
    "rdkit": ("https://www.rdkit.org/docs", None),
    "openeye": ("https://docs.eyesopen.com/toolkits/python/", None),
    "mdtraj": ("https://www.mdtraj.org/1.9.5/", None),
    "openff.interchange": (
        "https://docs.openforcefield.org/projects/interchange/en/stable/",
        None,
    ),
    "openff.fragmenter": (
        "https://docs.openforcefield.org/projects/fragmenter/en/stable/",
        None,
    ),
    "openff.units": (
        "https://docs.openforcefield.org/projects/units/en/stable/",
        None,
    ),
    "openff.nagl": (
        "https://docs.openforcefield.org/projects/nagl/en/stable/",
        None,
    ),
    "openff.docs": (
        "https://docs.openforcefield.org/en/latest/",
        None,
    ),
}
myst_url_schemes = [
    "http",
    "https",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# Extensions for the myst parser
myst_enable_extensions = [
    "dollarmath",
    "colon_fence",
    "smartquotes",
    "replacements",
    "deflist",
]
myst_heading_anchors = 4

# Myst NB settings
# Never execute notebooks - this should be done by CI
# Output is stored in the notebook itself
# Remember `Widgets -> Save Notebook Widget State` in the notebook.
nb_execution_mode = "off"

# sphinx-notfound-page
# https://github.com/readthedocs/sphinx-notfound-page
# Renders a 404 page with absolute links
if find_import_spec("notfound"):
    extensions.append("notfound.extension")

    notfound_urls_prefix = "/projects/toolkit/en/stable/"
    notfound_context = {
        "title": "404: File Not Found",
        "body": f"""
    <h1>404: File Not Found</h1>
    <p>
        Sorry, we couldn't find that page. This often happens as a result of
        following an outdated link. Please check the
        <a href="{notfound_urls_prefix}">latest stable version</a>
        of the docs, unless you're sure you want an earlier version, and
        try using the search box or the navigation menu on the left.
    </p>
    <p>
    </p>
    """,
    }

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
source_suffix = [".rst", ".md", ".ipynb"]

# The master toctree document.
master_doc = "index"

# General information about the project.
project = "OpenFF Toolkit"
copyright = "2016-2021 Open Force Field Initiative"
author = "Open Force Field Initiative"

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = openff.toolkit.__version__
# The full version, including alpha/beta/rc tags.
release = openff.toolkit.__version__

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = "en"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False


# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = "openff_sphinx_theme"

# (Optional) Logo and favicon
html_logo = "_static/images/logos/openff_toolkit_v1_white.png"
html_favicon = "_static/images/favicon.svg"

# Theme options are theme-specific and customize the look and feel of a
# theme further.
html_theme_options = {
    # Repository integration
    # Set the repo url for the link to appear
    "repo_url": "https://github.com/openforcefield/openff-toolkit",
    # The name of the repo. If must be set if repo_url is set
    "repo_name": "openff-toolkit",
    # Must be one of github, gitlab or bitbucket
    "repo_type": "github",
    # Colour for sidebar captions and other accents. One of
    # openff-toolkit-blue, openff-dataset-yellow, openff-evaluator-orange,
    # red, pink, purple, deep-purple, indigo, blue, light-blue, cyan,
    # teal, green, light-green, lime, yellow, amber, orange, deep-orange
    "color_accent": "openff-toolkit-blue",
    # Content Minification for deployment, prettification for debugging
    "html_minify": False,
    "html_prettify": False,
    "css_minify": True,
    "master_doc": False,
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
html_sidebars = {
    "**": ["globaltoc.html", "searchbox.html", "localtoc.html"],
}


# -- Options for HTMLHelp output ------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = "openforcefielddoc"


# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    "papersize": "letterpaper",
    # The font size ('10pt', '11pt' or '12pt').
    #
    "pointsize": "10pt",
    # Additional stuff for the LaTeX preamble.
    "preamble": r"""
        \usepackage{charter}
        \usepackage[defaultsans]{lato}
        \usepackage{inconsolata}
    """,
    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (
        master_doc,
        "openforcefield.tex",
        "OpenFF Toolkit Documentation",
        "Open Force Field Consortium",
        "manual",
    ),
]


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [(master_doc, "openff-toolkit", "OpenFF Toolkit Documentation", [author], 1)]


# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        master_doc,
        "openff-toolkit",
        "OpenFF Toolkit Documentation",
        author,
        "openff-toolkit",
        (
            "A modern, extensible library for molecular mechanics force field science from the Open Force "
            "Field Consortium.",
        ),
        "Miscellaneous",
    ),
]



import importlib
import pathlib
import warnings
from typing import TYPE_CHECKING

from openff.toolkit import Quantity, unit
from openff.toolkit.utils.base_wrapper import ToolkitWrapper
from openff.toolkit.utils.exceptions import (
    ToolkitUnavailableException,
)

if TYPE_CHECKING:
    from openff.toolkit.topology.molecule import FrozenMolecule, Molecule


__all__ = ("NAGLToolkitWrapper",)


class NAGLToolkitWrapper(ToolkitWrapper):
    """NAGL toolkit wrapper for applying partial charges with a GCN model.

    :external+openff.nagl:doc:`index` computes partial charges directly from the
    molecular graph and independent of conformer coordinates using a Graph
    Convolutional Network."""

    _toolkit_name = "OpenFF NAGL"
    _toolkit_installation_instructions = (
        "See https://docs.openforcefield.org/projects/nagl/en/latest/installation.html"
    )
    try:
        from openff.nagl_models import list_available_nagl_models

        _supported_charge_methods = {pathlib.Path(path).name: dict() for path in list_available_nagl_models()}
    except ImportError:
        _supported_charge_methods = dict()

    def __init__(self):
        super().__init__()

        if not self.is_available():
            raise ToolkitUnavailableException(
                f"The required toolkit {self._toolkit_name} is not "
                f"available. {self._toolkit_installation_instructions}"
            )
        else:
            from openff.nagl import __version__ as nagl_version

            self._toolkit_version = nagl_version

    @classmethod
    def is_available(cls) -> bool:
        if cls._is_available is None:
            try:
                importlib.import_module("openff.nagl")
            except ImportError:
                cls._is_available = False
            else:
                cls._is_available = True
        return cls._is_available

    def assign_partial_charges(
        self,
        molecule: "Molecule",
        partial_charge_method: str,
        use_conformers: list["Quantity"] | None = None,
        strict_n_conformers: bool = False,
        normalize_partial_charges: bool = True,
        doi: str | None = None,
        file_hash: str | None = None,
        _cls: type["FrozenMolecule"] | None = None,
    ):
        """
        Compute partial charges with NAGL and store in ``self.partial_charges``

        .. warning :: This API is experimental and subject to change.

        Parameters
        ----------
        molecule
            Molecule for which partial charges are to be computed
        partial_charge_method
            The NAGL model to use. May be a path or the name of a model in a
            directory from the ``openforcefield.nagl_model_path`` entry point.
        use_conformers
            This argument is ignored as NAGL does not generate or consider
            coordinates during inference.
        strict_n_conformers
            This argument is ignored as NAGL does not generate or consider
            coordinates during inference.
        normalize_partial_charges : bool, default=True
            Whether to offset partial charges so that they sum to the total
            formal charge of the molecule. This is used to prevent accumulation
            of rounding errors when the partial charge generation method has
            low precision.
        doi
            Zenodo DOI to check if NAGL model file needs to be fetched. Passed
            directly to openff.nagl_models._dynamic_fetch.get_model, see docs
            on that method for more details.
        file_hash
            sha256 hash to check against NAGL model file. Passed
            directly to openff.nagl_models._dynamic_fetch.get_model, see docs
            on that method for more details.
        _cls : class
            Molecule constructor

        Raises
        ------
        ChargeMethodUnavailableError
            if the requested charge method can not be handled by this toolkit

        ChargeCalculationError
            if the charge method is supported by this toolkit, but fails
        """
        from openff.nagl import GNNModel
        from openff.nagl_models._dynamic_fetch import get_model

        if partial_charge_method == "" or partial_charge_method == "None":
            raise FileNotFoundError(
                "NAGLToolkitWrapper.assign_partial_charges can not accept "
                "a blank model file name. There is no default model, one must be "
                "explicitly defined when being called."
            )

        if _cls is None:
            from openff.toolkit.topology.molecule import Molecule

            _cls = Molecule

        if use_conformers:
            warnings.warn(
                "`NAGLToolkitWrapper.assign_partial_charges` was passed optional argument "
                "`use_conformers` which will not be used. OpenFF NAGL does not generate "
                "conformers as part of assigning partial charges.",
                UserWarning,
                stacklevel=2,
            )

        if strict_n_conformers:
            warnings.warn(
                "`NAGLToolkitWrapper.assign_partial_charges` was passed optional argument "
                "`strict_n_conformers` which will not be used. OpenFF NAGL does not generate "
                "conformers as part of assigning partial charges.",
                UserWarning,
                stacklevel=2,
            )

        model_path = get_model(filename=partial_charge_method, doi=doi, file_hash=file_hash)

        model = GNNModel.load(model_path, eval_mode=True)
        charges = model.compute_property(
            molecule,
            as_numpy=True,
            readout_name="am1bcc_charges",
            check_domains=True,
            error_if_unsupported=True,
        )

        molecule.partial_charges = Quantity(
            charges.astype(float),
            unit.elementary_charge,
        )

        if normalize_partial_charges:
            molecule._normalize_partial_charges()



"""
This is a configuration file for PyTest.

This file is designed to be used with nbval: https://nbval.readthedocs.io/

The nbval-sanitize.cfg file is also for use with nbval.
"""


def pytest_collectstart(collector):
    """
    Ignore certain outputs when validating documentation notebooks

    See https://nbval.readthedocs.io/en/latest/#Skipping-certain-output-types
    """
    if collector.fspath and collector.fspath.ext == ".ipynb":
        collector.skip_compare += (
            # SVGs are often produced by Molecule.visualize()
            # The precise SVG output varies with environment and RDkit version,
            # and failures of visualize() produce exceptions or text output, so
            # we don't lose much by ignoring the actual content.
            "image/svg+xml",
        )



"""
Tests for utility methods

"""

import os

import pytest

from openff.toolkit import unit


def test_subclasses():
    """Test that all subclasses (and descendents) are correctly identified by all_subclasses()"""

    from openff.toolkit.utils.utils import all_subclasses

    class Foo:
        pass

    class FooSubclass1(Foo):
        pass

    class FooSubclass2(Foo):
        pass

    class FooSubSubclass(FooSubclass1):
        pass

    subclass_names = [cls.__name__ for cls in all_subclasses(Foo)]
    assert set(subclass_names) == set(["FooSubclass1", "FooSubclass2", "FooSubSubclass"])


def test_temporary_cd():
    """Test temporary_cd() context manager"""
    initial_dir = os.getcwd()
    temporary_dir = "/"
    from openff.toolkit.utils import temporary_cd

    with temporary_cd(temporary_dir):
        assert os.getcwd() == temporary_dir
    assert os.getcwd() == initial_dir


def test_get_data_file_path():
    """Test get_data_file_path()"""
    from openff.toolkit.utils import get_data_file_path

    filename = get_data_file_path("test_forcefields/tip3p.offxml")
    assert os.path.exists(filename)


def test_get_data_file_path_does_not_exist():
    from openff.toolkit.utils import get_data_file_path

    with pytest.raises(ValueError, match="does not exist"):
        get_data_file_path("bogus.offxml")


def test_dimensionless_units():
    from openff.toolkit.utils.utils import string_to_unit, unit_to_string

    assert string_to_unit("dimensionless") == unit.dimensionless

    unit_string = unit_to_string(unit.dimensionless)
    unit_value = string_to_unit(unit_string)

    assert unit_value == unit.dimensionless


def test_object_to_quantity_accepts_openmm():
    import openmm

    from openff.toolkit.utils.utils import object_to_quantity

    val = object_to_quantity(1.0 * openmm.unit.angstrom)
    assert val == 1.0 * unit.angstrom
    val = object_to_quantity(1.0 * openmm.unit.nanometer)
    assert val == 10.0 * unit.angstrom
    val = object_to_quantity(
        [
            2.0 * openmm.unit.angstrom,
            2.0 * openmm.unit.nanometer,
            2.0 * openmm.unit.dimensionless,
        ]
    )
    assert val == [2.0 * unit.angstrom, 20.0 * unit.angstrom, 2 * unit.dimensionless]


def test_sort_smirnoff_dict():
    from collections import OrderedDict

    from openff.toolkit.typing.engines.smirnoff import ForceField
    from openff.toolkit.utils.utils import get_data_file_path, sort_smirnoff_dict

    forcefield = ForceField(get_data_file_path("test_forcefields/test_forcefield.offxml"))
    smirnoff_dict = forcefield._to_smirnoff_data()

    # Ensure data is not created or destroyed
    # dict.__eq__ does not check order
    assert smirnoff_dict == OrderedDict(sort_smirnoff_dict(forcefield._to_smirnoff_data()))



"""This module defines functions for loading parameter handler and parser classes which
have been registered through the `entrypoint plugin system <https://packaging.python.
org/guides/creating-and-discovering-plugins/#using-package-metadata>`_.

.. warning ::

   This feature is experimental and may be removed / altered in future versions.

Currently only ``ParameterHandler`` classes can be registered via the plugin
system. This is possible by registering the handler class through an entry
point in your projects ``setup.py`` file::

    setup(
        ...
        entry_points={
            'openff.toolkit.plugins.handlers': ['CustomHandler = myapp:CustomHandler']
        },
        ...
    )

where in this example your package is named ``myapp`` and contains a class which
inherits from ``ParameterHandler`` named ``CustomHandler``.
"""

import logging
from importlib.metadata import entry_points

from openff.toolkit.typing.engines.smirnoff.parameters import ParameterHandler

logger = logging.getLogger(__name__)

SUPPORTED_PLUGIN_NAMES = ["handlers"]  # io_handlers could be supported in future.


def _load_handler_plugins(handler_name: str, expected_type: type) -> list[type]:
    """Loads parameter handler plugins of a specified type which have been registered
    with the ``entrypoint`` plugin system.

    Parameters
    ----------
    handler_name
        The name of the hander plugin. This can currently be any of the names
        listed in ``SUPPORTED_PLUGIN_NAMES``.
    expected_type
        The expected class type of the plugin. E.g. when loading parameter io
        handler plugins the expected class type is ``ParameterIOHandler``. Any
        classes not matching the expected type will be skipped.
    """

    if handler_name not in SUPPORTED_PLUGIN_NAMES:
        raise NotImplementedError(
            f"Plugins named {handler_name} not supported. The list of currently "
            f"supported plugin names is: {SUPPORTED_PLUGIN_NAMES}."
        )

    PluginType = type[ParameterHandler]

    discovered_plugins: list[PluginType] = list()

    for entry_point in entry_points().select(group=f"openff.toolkit.plugins.{handler_name}"):
        try:
            discovered_plugins.append(entry_point.load())
        except (ImportError, AttributeError):
            logger.exception(f"Could not load the {entry_point} plugin")
            continue

    valid_plugins: list[PluginType] = list()

    for discovered_plugin in discovered_plugins:
        if not issubclass(discovered_plugin, expected_type):
            logger.info(
                f"The {discovered_plugin.__name__} object has been registered as a "
                f"{handler_name} plugin, but does not inherit from "
                f"{expected_type.__name__}. This plugin will be skipped."
            )
            continue

        valid_plugins.append(discovered_plugin)

    return valid_plugins


def load_handler_plugins() -> list[type[ParameterHandler]]:
    """Loads any ``ParameterHandler`` class plugins which have been registered through
    the ``entrypoint`` plugin system.

    Returns
    -------
        The registered ``ParameterHandler`` plugins.
    """
    return _load_handler_plugins("handlers", ParameterHandler)



"""
Built-in ToolkitWrapper for very basic functionality. Intended for testing and not much more.
"""

__all__ = ("BuiltInToolkitWrapper",)

from typing import TYPE_CHECKING

from openff.toolkit import Quantity, unit
from openff.toolkit.utils import base_wrapper
from openff.toolkit.utils.exceptions import ChargeMethodUnavailableError
from openff.toolkit.utils.utils import inherit_docstrings

if TYPE_CHECKING:
    from openff.toolkit.topology.molecule import FrozenMolecule


@inherit_docstrings
class BuiltInToolkitWrapper(base_wrapper.ToolkitWrapper):
    """
    Built-in ToolkitWrapper for very basic functionality. Intended for testing and not much more.

    .. warning :: This API is experimental and subject to change.
    """

    _toolkit_name = "Built-in Toolkit"
    _toolkit_installation_instructions = (
        "This toolkit is installed with the Open Force Field Toolkit and does not require additional dependencies."
    )
    _supported_charge_methods = {
        "zeros": {"rec_confs": 0, "min_confs": 0, "max_confs": 0},
        "formal_charge": {"rec_confs": 0, "min_confs": 0, "max_confs": 0},
    }

    PARTIAL_CHARGE_METHODS = _supported_charge_methods

    def __init__(self):
        super().__init__()

    def assign_partial_charges(
        self,
        molecule: "FrozenMolecule",
        partial_charge_method: str | None = None,
        use_conformers: Quantity | None = None,
        strict_n_conformers: bool = False,
        normalize_partial_charges: bool = True,
        _cls: type | None = None,
    ):
        """

        Compute partial charges with the built-in toolkit using simple arithmetic operations,
        and assign the new values to the partial_charges attribute.


        .. warning :: This API is experimental and subject to change.

        Parameters
        ----------
        molecule
            Molecule for which partial charges are to be computed
        partial_charge_method
            The charge model to use. One of ['zeros', 'formal_charge']. If None, 'formal_charge'
            will be used.
        use_conformers
            (n_atoms, 3) and dimension of distance. Optional, default = None
            Coordinates to use for partial charge calculation. If None, an appropriate number
            of conformers will be generated.
        strict_n_conformers
            Whether to raise an exception if an invalid number of conformers is provided for the
            given charge method.
            If this is False and an invalid number of conformers is found, a warning will be raised
            instead of an Exception.
        normalize_partial_charges
            Whether to offset partial charges so that they sum to the total formal charge of the molecule.
            This is used to prevent accumulation of rounding errors when the partial charge generation method has
            low precision.
        _cls
            Molecule constructor

        Raises
        ------
        ChargeMethodUnavailableError if this toolkit cannot handle the requested charge method

        IncorrectNumConformersError if strict_n_conformers is True and use_conformers is provided
        and specifies an invalid number of conformers for the requested method

        ChargeCalculationError if the charge calculation is supported by this toolkit, but fails
        """

        if partial_charge_method is None:
            partial_charge_method = "formal_charge"

        if _cls is None:
            from openff.toolkit.topology.molecule import Molecule

            _cls = Molecule

        # Make a temporary copy of the molecule, since we'll be messing with its conformers
        mol_copy = _cls(molecule)

        partial_charge_method = partial_charge_method.lower()
        if partial_charge_method not in self._supported_charge_methods:
            raise ChargeMethodUnavailableError(
                f'Partial charge method "{partial_charge_method}"" is not supported by '
                f"the Built-in toolkit. Available charge methods are "
                f"{self._supported_charge_methods}"
            )

        if use_conformers is None:
            # Note that this refers back to the GLOBAL_TOOLKIT_REGISTRY by default, since
            # BuiltInToolkitWrapper can't generate conformers
            mol_copy.generate_conformers(
                n_conformers=self._supported_charge_methods[partial_charge_method]["rec_confs"]
            )
        else:
            mol_copy._conformers = None
            for conformer in use_conformers:  # type: ignore[attr-defined]
                mol_copy._add_conformer(conformer)
            self._check_n_conformers(
                mol_copy,
                partial_charge_method=partial_charge_method,
                min_confs=0,
                max_confs=0,
                strict_n_conformers=strict_n_conformers,
            )

        if partial_charge_method == "zeros":
            partial_charges = [0.0] * molecule.n_atoms

        elif partial_charge_method == "formal_charge":
            partial_charges = [float(atom.formal_charge.m) for atom in molecule.atoms]  # type: ignore

        molecule.partial_charges = Quantity(partial_charges, unit.elementary_charge)

        if normalize_partial_charges:
            molecule._normalize_partial_charges()
